
class AsyncClass(object):
    def __init__(self):
        pass

    def __await__(self):
        async def closure():
            return self

        return closure().__await__()

    async def __aenter__(self):
        return self

    async def __aexit__(self, *args):
        pass

    async def initialize(self):
        raise NotImplementedError("Subclass should override")



from html.parser import HTMLParser


# https://docs.python.org/3/library/html.parser.html#examples
class MyHTMLParser(HTMLParser):
    def __init__(self):
        super().__init__()
        self.content = ""
        self.a_exist = False

    def handle_starttag(self, tag, attrs):
        self.a_exist = True

    def handle_data(self, data):
        self.content += " " + data



from abc import ABC, abstractmethod
from typing import Dict

import requests
from cssselect import ExpressionError
from goofis_ardihikaru.utils.async_class import AsyncClass
from goofis_ardihikaru.utils.currency import clean_currency, get_percent_value
from goofis_ardihikaru.utils.user_agent import build_fake_ua
from goofis_ardihikaru.utils.utils import reformat_name
from parsel import Selector
from requests import Response


class AbcParser(AsyncClass, ABC):
    def __init__(self, code: str, lang: str, timeout: int = 30, url: str = "https://www.google.com/finance/quote",
                 headers: Dict = None):
        super().__init__()
        if headers is None:
            headers = {"User-Agent": build_fake_ua()}
        self.lang = lang
        self.timeout = timeout
        self.url = f"{url}/{code}?hl={lang}"

        self.headers = headers
        if headers is None:
            self.headers = {
                "User-Agent": build_fake_ua()
            }

        self.i = 0
        self.j = 0
        self.info = None

    def incr(self) -> int:
        current_idx = self.i

        # do increment
        self.i += 1

        return current_idx

    def incr_j(self) -> int:
        current_idx = self.j

        # do increment
        self.j += 1

        return current_idx

    async def get_html_text(self) -> str:
        html = await self.get_html()

        return html.text

    async def get(self):
        html = await self.get_html()

        return await self.parser(html)

    async def get_html(self):
        cookies = {"CONSENT": "YES+cb.20210720-07-p0.en+FX+410"}
        # return requests.get(self.url, headers=self.headers, timeout=self.timeout, cookies={'CONSENT': 'YES+'})
        return requests.get(self.url, headers=self.headers, timeout=self.timeout, cookies=cookies)

    @abstractmethod
    async def parser(self, html: Response):
        pass

    @staticmethod
    async def fetch_current_price(selector: Selector, lang: str) -> float:
        """
        fetches current price
        """
        try:
            current_price_str = selector.css(".YMlKec.fxKbKc::text").get()
            return clean_currency(current_price_str, lang)
        except ExpressionError:
            return 0.0

    @staticmethod
    async def fetch_current_price_percent(selector: Selector, lang: str) -> float:
        """
        fetches current price
        """
        try:
            current_price_percent_str = selector.css(".JwB6zf::text").get()
            return get_percent_value(current_price_percent_str, lang)
        except ExpressionError:
            return 0.0

    async def fetch_current_price_percent_v2(self) -> float:
        """
        fetches current price percent changes
        """
        try:
            # current price
            current_price = self.info.current_price
            yesterday_price = self.info.general.prev_price_close

            diff = yesterday_price - current_price
            minus = 1
            if diff < -1:
                minus = -1

            # to percent
            diff_percent = diff / yesterday_price * 100
            diff_percent = round((diff_percent * minus), 2)

            return diff_percent
        except Exception:
            return 0.0

    @staticmethod
    async def fetch_current_price_value(last_closed_price: float, current_price: float) -> float:
        """
        fetches current price value
        """
        try:
            return round((last_closed_price - current_price), 2)
        except ExpressionError:
            return 0.0

    async def fetch_current_price_value_v2(self) -> float:
        """
        fetches current price changes
        """
        try:
            # current price
            current_price = self.info.current_price
            yesterday_price = self.info.general.prev_price_close

            diff = yesterday_price - current_price
            minus = 1
            if diff < -1:
                minus = -1
            diff = round((diff * minus), 2)

            return diff
        except Exception:
            return 0.0

    @staticmethod
    async def fetch_about_info(selector: Selector) -> str:
        """
        fetches `about company` information

        one example with missing `about company` data: BCIP:IDX
        """
        about = ""
        try:
            about = selector.css(".bLLb2d::text").get()
        except TypeError:
            pass
        except Exception:
            pass

        return about

    @staticmethod
    async def fetch_name(selector: Selector, about: str, lang: str) -> str:
        """
        fetches `name` information
        """
        name = ""
        try:
            name = selector.css(".zzDege::text").get()
            if name is None:
                raise TypeError("`name` is None")
        except TypeError:
            pass
        except Exception:
            pass

        # optional: casts value by collecting value from the `about company`
        name = reformat_name(name, about, lang)

        return name.strip()



from dataclasses import dataclass, asdict
from typing import Dict, Union, Any

import simplejson as json

from goofis_ardihikaru.enums.price_change import ValueChange


@dataclass
class General:
    """ General Data model """

    prev_price_close: float = 0.0

    day_range_min: float = 0.0
    day_range_max: float = 0.0
    day_range_diff: float = 0.0
    day_range_increase: bool = False

    year_range_min: float = 0.0
    year_range_max: float = 0.0
    year_range_diff: float = 0.0
    year_range_increase: bool = False

    # 1 trillion (T) = 1,000 billions (B)
    market_cap: float = 0.0
    market_cap_unit: float = 0.0

    avg_volume: float = 0.0
    avg_volume_unit: float = 0.0

    pe_ratio: float = 0.0

    dividend_yield: float = 0.0
    primary_exchange: str = ""

    ceo: str = ""
    founded_at: str = ""
    website: str = ""
    head_office: str = ""
    employees: int = 0


@dataclass
class FinancialValue:
    """ Percentage Price Data model """

    unit: str = ""
    value: float = 0.0
    percent: float = 0.0
    price_change: str = ValueChange.NONE.value
    percent_change: str = ValueChange.NONE.value
    empty: bool = False
    empty_percent: bool = False


@dataclass
class IncomeStatement:
    """ Income Statement Data model """

    revenue: FinancialValue = FinancialValue()
    operating_expense: FinancialValue = FinancialValue()
    net_income: FinancialValue = FinancialValue()
    net_income_margin: FinancialValue = FinancialValue()
    earning_per_share: FinancialValue = FinancialValue()
    ebitda: FinancialValue = FinancialValue()
    effective_tax_rate: FinancialValue = FinancialValue()


@dataclass
class BalanceSheet:
    """ Balance Sheet Data model """

    cash_n_short_term_investment: FinancialValue = FinancialValue()
    total_assets: FinancialValue = FinancialValue()
    total_liabilities: FinancialValue = FinancialValue()
    total_equity: FinancialValue = FinancialValue()
    share_outstanding: FinancialValue = FinancialValue()
    price_to_book: FinancialValue = FinancialValue()
    return_on_assets: FinancialValue = FinancialValue()
    return_on_capital: FinancialValue = FinancialValue()


@dataclass
class CashFlow:
    """ Cash Flow Data model """

    net_income: FinancialValue = FinancialValue()
    cash_from_operation: FinancialValue = FinancialValue()
    cash_from_investing: FinancialValue = FinancialValue()
    cash_from_financing: FinancialValue = FinancialValue()
    net_change_in_cash: FinancialValue = FinancialValue()
    free_cash_flow: FinancialValue = FinancialValue()


@dataclass
class FinancialCategory:
    """ Financial Category Data model """

    income_statement: IncomeStatement = IncomeStatement()
    balance_sheet: BalanceSheet = BalanceSheet()
    cash_flow: CashFlow = CashFlow()


@dataclass
class Financial:
    """ Financial Data model """

    # quarter data
    quarter: FinancialCategory = FinancialCategory()

    # TODO: add annual data


@dataclass
class StockData:
    """ Stock data model """

    FN_ID = "id"
    FN_GENERAL = "general"
    FN_ABOUT = "about"
    FN_FINANCIAL = "financial"
    FN_CURRENT_PRICE = "current_price"
    FN_CURRENT_PRICE_CHANGE_PERCENT = "current_price_change_percent"
    FN_CURRENT_PRICE_CHANGE_VALUE = "current_price_change_value"

    id: str = "-1"
    general: General = General()
    financial: Financial = Financial()
    name: str = ""
    about: str = ""
    current_price: float = 0.0
    current_price_change_percent: float = 0.0
    current_price_change_value: float = 0.0

    @staticmethod
    def sanitize(val: str) -> str:
        """ sanitizes json string """
        val = val.replace("'", '"')
        val = val.replace("False", 'false')
        val = val.replace("True", 'true')

        return val

    def val_to_dict(self, val: Any) -> Union[None, int, str, Dict]:
        if val is None:
            return None
        if isinstance(val, int):
            return val
        if isinstance(val, str):
            return val

        val = str(val)

        # sanitizes
        val = self.sanitize(val)

        return json.loads(self.sanitize(val))

    def to_dict(self, with_id: bool = False):
        dict_data = asdict(self)

        # remove id if disabled
        if not with_id:
            dict_data.pop(self.FN_ID)

        return {
            k: self.val_to_dict(v) for k, v in dict_data.items()
        }



from typing import List, Tuple

from goofis_ardihikaru.dto.stock_data import FinancialValue
from goofis_ardihikaru.enums.currency_unit import CurrencyUnit
from goofis_ardihikaru.enums.language import Language
from goofis_ardihikaru.enums.price_change import ValueChange


def clean_currency(currency: str, lang: str) -> float:
    if currency == "-":
        return 0.0

    # sometimes it contains a percent symbol
    currency = currency.replace("%", "")

    currency = currency.replace("Rp", "")

    # for Indonesian-based format (?hl=id)
    # WARNING: kalau indonesia (?hl=id), delimiter ribuan pakai titik!
    if lang == Language.INDONESIA.value:
        currency = currency.replace(".00", "")
        currency = currency.replace(".", "")
        currency = currency.replace(",", ".")
    else:
        currency = currency.replace(",00", "")
        currency = currency.replace(".", "")

    currency = currency.replace(" ", "")

    return float(currency)


def get_price_ranges(price_range: str, lang: str) -> Tuple[float, float, float]:
    # first, split
    price_range_arr: List = price_range.split(" - ")

    # then, captures the data
    price_min: float = clean_currency(price_range_arr[0], lang)
    price_max: float = clean_currency(price_range_arr[1], lang)
    price_diff: float = round((price_max - price_min), 2)

    return price_min, price_max, price_diff


def get_price_with_unit(price_with_unit: str, lang: str) -> Tuple[float, str]:
    # sanitize
    price_with_unit = price_with_unit.replace("\xa0", "")  # removes space

    if f" {CurrencyUnit.ID_JT.value}" in price_with_unit:
        price_with_unit = price_with_unit.replace(" jt", " jt")

    # first, split
    price_range_arr: List = price_with_unit.split(" ")

    # only capture price and remove the currency symbol
    price_str = price_range_arr[0]

    # extracts data
    # unit `jt` is only for lang=id (Indonesia)
    if CurrencyUnit.ID_JT.value in price_str or CurrencyUnit.ID_RB.value in price_str:
        market_cap_unit = price_str[-2:]
        price = clean_currency(price_str[:-2], lang)
    else:
        market_cap_unit = price_str[-1]
        price = clean_currency(price_str[:-1], lang)

    return price, market_cap_unit


def get_percent_value(percent_val: str, lang: str) -> float:
    if percent_val == "-" or percent_val == "—":
        return 0.0

    # only for lang=id (Indonesia)
    if lang == Language.INDONESIA.value:
        percent_val = percent_val.replace(".", "")
        percent_val = percent_val.replace(",", ".")

    percent_val = percent_val.replace("%", "")
    percent_val = percent_val.replace("+", "")

    return float(percent_val)


def contains_unit(val: str) -> bool:
    """
        checks if it contains unit (e.g. `B`, `T`, `M`) or not
    """
    last_char = val[-1]

    try:
        # if int found, it DOES not contain unit
        _ = int(last_char)
        return False
    except ValueError:
        return True


def get_financial_value(lang: str, value_str: str, percent_str: str) -> FinancialValue:
    # empty value, ignore it
    if value_str == "—" and percent_str == "—":
        return FinancialValue(empty=True)
    elif value_str == "∞" and percent_str == "—":
        return FinancialValue(empty=True)

    # extracts data
    if contains_unit(value_str):
        # unit `jt` is only for lang=id (Indonesia)
        if CurrencyUnit.ID_JT.value in value_str or CurrencyUnit.ID_RB.value in value_str:
            value_unit = value_str[-2]
            value = clean_currency(value_str[:-2], lang)
        else:
            value_unit = value_str[-1]
            value = clean_currency(value_str[:-1], lang)
    else:
        value_unit = ""
        value = clean_currency(value_str, lang)

    empty_percent = False
    if percent_str == "—":
        empty_percent = True

    percent: float = get_percent_value(percent_str, lang)

    # builds percent change value
    percent_change: str = ValueChange.NONE.value
    if percent == "—":
        percent_change = ValueChange.NONE.value
    elif percent > 0:
        percent_change = ValueChange.INCREASE.value
    elif percent < 0:
        percent_change = ValueChange.DECREASE.value

    # builds price change value
    price_change: str = ValueChange.NONE.value
    if value == "—":
        price_change = ValueChange.INCREASE.value
    elif value > 0:
        price_change = ValueChange.INCREASE.value
    elif value < 0:
        price_change = ValueChange.DECREASE.value

    return FinancialValue(
        unit=value_unit,
        value=value,
        percent=percent,
        price_change=price_change,
        percent_change=percent_change,
        empty_percent=empty_percent,
    )



from goofis_ardihikaru.enums.language import Language


def reformat_name(name: str, about: str, lang: str) -> str:
    # do nothing if this logic below got triggered
    if about is None or len(about) < 0 or len(name) < 0:
        return name
    if lang != Language.INDONESIA.value:
        return name

    # splits DOT first
    about_arr = about.split(".")
    about = about_arr[0]

    # starts converting value
    about_short = about[0:120]  # gets first 120 chars
    if "merupakan" in about_short:
        about_arr = about_short.split("merupakan")
        name = about_arr[0]

    elif "adalah" in about[0:120]:
        about_arr = about_short.split("adalah")
        name = about_arr[0]

    if "adalah" in name:
        about_arr = about_short.split("adalah")
        name = about_arr[0]

    # makes it shorter further
    if "atau" in name:
        about_arr = about_short.split("atau")
        name = about_arr[0]

    # makes it shorter further
    if "disingkat" in about_short:
        about_arr = about_short.split("disingkat")
        name = about_arr[0]

    # remove dots if exists
    name_arr = name.split(".")
    name = name_arr[0]

    name = name.replace("Perusahaan Perseroan ", "")

    return name.strip()



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="goofis-ardihikaru",
    version="0.0.48",
    license='MIT',
    author="Muhammad Febrian Ardiansyah",
    author_email="mfardiansyah@outlook.com",
    description="Modules to extract information from Google Finance",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://pypi.org/project/goofis-ardihikaru/0.0.1/#files",
    install_requires=[  # package dependency
        'ujson==5.8.0',
        'requests==2.31.0',
        'parsel==1.8.1',
        'fake-useragent==1.2.1',
        'simplejson==3.19.1',
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    platforms='any',
    keywords='asyncio',
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.8"
)
